"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from colorama import Fore, Style


def color_character_names(string: str) -> str:
    """
    Colors character names in their respective house colors
    :param string: The string to format
    :return: The formatted string
    """
    for color, house in [
        (Fore.LIGHTRED_EX, [
            "Edelgard",
            "Hubert",
            "Ferdinand",
            "Linhardt",
            "Caspar",
            "Petra",
            "Bernadetta",
            "Dorothea"
        ]),
        (Fore.LIGHTBLUE_EX, [
            "Dimitri",
            "Dedue",
            "Felix",
            "Sylvain",
            "Ashe",
            "Annette",
            "Mercedes",
            "Ingrid"
        ]),
        (Fore.LIGHTYELLOW_EX, [
            "Claude",
            "Hilda",
            "Lorenz",
            "Raphael",
            "Ignatz",
            "Lysithea",
            "Marianne",
            "Leonie"
        ]),
        (Fore.LIGHTGREEN_EX, [
            "Rhea",
            "Sothis",
            "Seteth",
            "Flayn",
            "Hanneman",
            "Manuela",
            "Alois",
            "Gilbert",
            "Catherine",
            "Shamir",
            "Cyril"
        ]),
        (Fore.LIGHTWHITE_EX, [
            "Byleth M",
            "Byleth F"
        ]),
    ]:
        for character in house:
            string = string.replace(character, "{}{}{}".format(
                color, character, Style.RESET_ALL
            ))
    return string



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Tuple


def calculate_remaining_support_levels(
        character: str,
        support_levels: Dict[str, Dict[str, str]],
        support_levels_info: Dict[str, Dict[str, str]]
) -> Tuple[int, Dict[str, int]]:
    """
    Calculates how many support levels are left for a character.
    This calculates both the total amount of support levels left for the
    character as well as the remaining support levels with individual
    characters.
    :param character: The character for which to calculate the remaining
                      support levels
    :param support_levels: The current support levels
    :param support_levels_info: The support levels info
    :return: A tuple consisting of the total remaining support levels and
             a dictionary mapping individual support levels to characters
    """
    total = 0
    remaining = {}

    for support, info in support_levels_info[character].items():
        current = support_levels[character][support]
        try:
            position = info.index(current)
            diff = len(info) - position - 1

            # TODO REMOVE ONCE DEVELOPMENT COMPLETE
            index = -1 if diff == 0 else -(diff + 1)
            calculated_rank = info[index]
            if current != calculated_rank:
                print("RANK ERROR!")

        except ValueError:
            diff = len(info)

        total += diff
        remaining[support] = diff

    return total, remaining


def calculate_max_support_rank(
        char_one: str,
        char_two: str,
        support_levels_info: Dict[str, Dict[str, str]]
) -> str:
    """
    Calculates the maximal support level between two characters
    :param char_one: The first character
    :param char_two: The second character
    :param support_levels_info: The support levels info
    :return: The maximal rank
    """
    return support_levels_info[char_one][char_two][-1]



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


def encode_support_levels(support_levels: str) -> str:
    """
    Converts a support level string into something that can easily be indexed.
    The existance of a '+' rank like 'C+' is stored in the info sheet as
    "CC" and therefore uses up two spaces, making indexing hard.
    So we convert these '+' support levels like this:
        C+ -> 1
        B+ -> 2
        A+ -> 3
    :param support_levels: The support level string to encode
    :return: The encoded support level string
    """
    return support_levels\
        .replace("C+", "1")\
        .replace("B+", "2")\
        .replace("A+", "3")\
        .replace("CC", "C1")\
        .replace("BB", "B2")\
        .replace("AA", "A3")


def decode_support_levels(encoded: str, for_sheet: bool = False):
    """
    Decodes the support levels back into a human-readable format
    :param encoded: The encoded support level
    :param for_sheet: Changes the support levels into the format used in the
                      data sheet (e.g. C1 -> CC)
    :return: The decoded support level
    """
    if for_sheet:
        return encoded \
            .replace("1", "C") \
            .replace("2", "B") \
            .replace("3", "A")
    else:
        return encoded\
            .replace("1", "C+")\
            .replace("2", "B+")\
            .replace("3", "A+")



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from fe3h_checklist.ods_interface import load_support_levels, \
    load_support_levels_info
from fe3h_checklist.calc import calculate_remaining_support_levels, \
    calculate_max_support_rank
from fe3h_checklist.support_encoding import decode_support_levels
from fe3h_checklist.formatting import color_character_names


def print_remaining_support_levels(no_byleth: bool, multi_line: bool):
    """
    Prints the remaining support levels for all characters and orders
    them in a descending order.
    :param no_byleth: Whether or not to include support levels with Byleth
    :param multi_line: Whether or not to use multiple lines per character
    :return: None
    """
    support_levels = load_support_levels()
    support_levels_info = load_support_levels_info()

    ranking = []
    for character in support_levels_info:

        if "Byleth" in character and no_byleth:
            continue

        total, remaining = calculate_remaining_support_levels(
            character, support_levels, support_levels_info
        )

        if "Byleth" not in character and no_byleth:
            total -= remaining.pop("Byleth M", 0)
            total -= remaining.pop("Byleth F", 0)

        ranking.append((character, total, remaining))

    ranking.sort(key=lambda x: x[1], reverse=True)
    max_char_length = len(max(ranking, key=lambda x: len(x[0]))[0])

    for rank, (character, total, remaining) in enumerate(ranking):

        supports = []
        sorted_remaining = []
        for support, diff in remaining.items():
            sorted_remaining.append((support, diff))
        sorted_remaining.sort(key=lambda x: x[1], reverse=True)

        for support, diff in sorted_remaining:

            if diff == 0:
                continue

            current_rank = support_levels[character][support]
            max_rank = calculate_max_support_rank(
                character, support, support_levels_info
            )

            supports.append("{}({}/{})".format(
                support.ljust(max_char_length) + (" " if multi_line else ""),
                decode_support_levels(current_rank),
                decode_support_levels(max_rank)
            ).ljust(max_char_length + 7))

        if multi_line:
            supports = ("\n" + " " * (max_char_length + 13)).join(supports)
        else:
            supports = "[" + ", ".join(supports) + "]"

        print(color_character_names("{}({}): {} -> {}".format(
            str(rank + 1).rjust(2),
            str(total).rjust(3),
            character.ljust(max_char_length),
            supports
        )))



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict
from pyexcel_ods import read_data, save_data
from fe3h_checklist import support_levels_file, support_levels_info_file
from fe3h_checklist.support_encoding import encode_support_levels, \
    decode_support_levels


def load_support_levels() -> Dict[str, Dict[str, str]]:
    """
    Loads the information about support levels from the ODS file
    and puts the data into a dictionary mapping the character's current
    support levels to each other
    :return: The dictionary mapping the support levels to characters.
             Example: {"Byleth M": {"Edelgard": "A"}}
    """
    return load_table(support_levels_file, "SupportLevels")


def load_support_levels_info() -> Dict[str, Dict[str, str]]:
    """
    Loads the information about support level info from the ODS file
    and puts the data into a dictionary mapping the character's support
    levels to each other
    :return: The dictionary mapping the support levels to characters.
             Example: {"Byleth M": {"Edelgard": "CBAS"}}
    """
    return load_table(support_levels_info_file, "SupportLevelsInfo")


def load_table(ods_file: str, sheet_name: str) -> Dict[str, Dict[str, str]]:
    """
    Loads a table from an ODS file.
    Assumptions about the ODS file:
        - Character names at the top and left side of the table
    :param ods_file: The ODS file from which to load the data
    :param sheet_name: The sheet name from which to load the data
    :return: A dictionary mapping the characters to each other and
             containing the value mapped to both of them.
             Example: {"Byleth M": {"Edelgard": "CBAS"}}
    """
    data_map = {}
    data_sheet = dict(read_data(ods_file))[sheet_name]

    names = data_sheet.pop(0)
    names.pop(0)

    for row in data_sheet:
        name = row.pop(0)
        data_map[name] = {}

        for i, entry in enumerate(row):
            if entry != "":
                data_map[name][names[i]] = encode_support_levels(entry)

    return data_map


def save_support_levels(support_levels: Dict[str, Dict[str, str]]):
    """
    Saves the support levels to file
    :param support_levels: The support levels to save
    :return: None
    """
    write_table(
        support_levels_file,
        "SupportLevels",
        support_levels,
        plus_notation=True
    )


def write_table(
        ods_file: str,
        sheet_name: str,
        data: Dict[str, Dict[str, str]],
        plus_notation: bool = False
):
    """
    Writes the content of a dictionary mapping characters to each other
    to an ODS file.
    :param ods_file: The ODS file to write to
    :param sheet_name: The sheet name to write to
    :param data: The data to write
    :param plus_notation: Whether or not to use +-notation.
                          True: 1 -> C+
                          False: C1 -> CC
    :return: None
    """
    new_sheet = [[""]]
    for char_one in data:
        new_sheet[0].append(char_one)
        new_row = [char_one]
        for char_two in data:
            entry = data[char_one].get(char_two, "")
            entry = decode_support_levels(entry, for_sheet=not plus_notation)
            new_row.append(entry)
        new_sheet.append(new_row)

    ods = dict(read_data(ods_file))
    ods[sheet_name] = new_sheet
    save_data(ods_file, ods)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import sys
from fe3h_checklist.support_encoding import encode_support_levels
from fe3h_checklist.ods_interface import load_support_levels_info, \
    load_support_levels, save_support_levels


def set_support_level(char_one: str, char_two: str, level: str):
    """
    Sets the support level for two characters and saves it.
    :param char_one: The first character
    :param char_two: The second character
    :param level: The support level
    :return: None
    """
    level = encode_support_levels(level)
    current = load_support_levels()
    info = load_support_levels_info()

    if char_two not in info[char_one]:
        print("Invalid character combination")
        sys.exit(1)
    elif level not in info[char_one][char_two] and level != "X":
        print("Invalid support level")
        sys.exit(1)
    else:
        current[char_one][char_two] = level
        current[char_two][char_one] = level
        save_support_levels(current)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fe3h-checklist.

fe3h-checklist is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fe3h-checklist is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fe3h-checklist.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os

local_dir = os.path.join(os.path.expanduser("~"), ".config/fe3h-checklist")
"""
Directory containing local config files
"""

support_levels_info_file = os.path.join(local_dir, "supportinfo.ods")
"""
File containing the support level info file
"""

support_levels_file = os.path.join(local_dir, "supportlevels.ods")
"""
File containing the support level info file
"""
